# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from __future__ import annotations

import hashlib
import json
import logging
import os
from pathlib import Path
from uuid import UUID

from majormode.perseus.utils import cast

from bootloader.ue.constant.ue_asset import ASSET_CLASS_NAME_MAPPING
from bootloader.ue.constant.ue_asset import UnrealEngineAssetClass


class UnrealEngineAsset:
    def __eq__(self, other: UnrealEngineAsset):
        """
        Check whether this asset is the same as another asset.

        Two assets are equivalent if they have the same name, the same class,
        the same package name, and the same file's checksum.


        :param other: Another asset.


        :return: ``True`` if the two assets are the same; ``False`` otherwise.
        """
        return other is not None \
            and self.__file_checksum == other.file_checksum \
            and self.__asset_class_path == other.asset_class_path \
            and self.__asset_name == other.asset_name \
            and self.__package_name == other.package_name

    def __init__(
            self,
            asset_name: str,
            asset_class_path: str,
            package_name: str,
            dependencies: list[str] or None,
            file_size: int,
            file_checksum: str,
            asset_id: UUID = None):
        """
        Build a new {@link UAsset}.


        :param asset_name: The name of the asset without the package.

        :param asset_class_path: The path name of the asset’s class.

        :param package_name: The name of the package in which the asset is
            found.

        :param dependencies: The list of package names of the assets that this
            asset depends on.

        :param file_size: The size in bytes of the asset file.

        :param file_checksum: The SHA256 message digest of the binary data of
            the asset file.

        :param asset_id: The identification of the asset, when already
            registered to the back-end platform.
        """
        self.__asset_id = asset_id
        self.__asset_name = asset_name
        self.__asset_class_path = asset_class_path
        self.__package_name = package_name
        self.__dependencies = dependencies
        self.__file_size = file_size
        self.__file_checksum = file_checksum

    @property
    def asset_class(self) -> UnrealEngineAssetClass:
        """
        Return the asset's class.

        For example, the class of a `/Script/Engine/SkeletalMesh` asset is
        `AssetClass.SkeletalMesh`


        :return: The class of the asset, or `None` if the class is not defined
            (i.e., a user-defined class).
        """
        class_name = self.__asset_class_path.split('/')[-1]
        try:
            asset_class = cast.string_to_enum(class_name, UnrealEngineAssetClass)
            return asset_class
        except ValueError as error:
            logging.exception(error)

    @property
    def asset_class_name(self) -> str:
        """
        Return the humanly readable name of the asset's class.

        For example, the humanly readable name of the class
        `/Script/Engine/SkeletalMesh` is `Skeletal Mesh`.


        :return: The humanly readable name of the asset's class.
        """
        return ASSET_CLASS_NAME_MAPPING.get(self.asset_class)

    @property
    def asset_class_path(self) -> str:
        """
        Return the path name of the asset's class.

        Examples:

        ```text
        /Script/Engine/SkeletalMesh
        /Script/Engine/Skeleton
        /Script/Engine/Texture2D
        ```

        :return: The path name of the asset's class.
        """
        return self.__asset_class_path

    @property
    def asset_id(self) -> UUID:
        if self.__asset_id is None:
            raise ValueError(f"No identification defined for the asset {self.__package_name}")

        return self.__asset_id

    @asset_id.setter
    def asset_id(self, asset_id: UUID):
        if self.__asset_id is not None:
            raise ValueError(f"The asset {self.__asset_name} has already an identification ({self.__asset_id})")

        self.__asset_id = asset_id

    @property
    def asset_name(self) -> str:
        """
        Return the name of the asset.


        :return: The name of the asset without the package.
        """
        return self.__asset_name

    @property
    def dependencies(self) -> list[str] or None:
        """
        Return the list of the package names of the assets that this asset
        depends on.


        :return: The list of package names of the assets that this asset
            depends on.
        """
        return self.__dependencies

    @property
    def file_checksum(self) -> str:
        return self.__file_checksum

    @property
    def file_size(self) -> int:
        return self.__file_size

    @staticmethod
    def from_json(payload: any) -> UnrealEngineAsset:
        return UnrealEngineAsset(
            payload['asset_name'],
            payload['asset_class_path'],
            payload['package_name'],
            payload.get('dependencies'),
            payload['file_size'],
            payload['file_checksum']
        )

    def to_json(self) -> any:
        return {
            'asset_name': self.__asset_name,
            'asset_class_path': self.__asset_class_path,
            'package_name': self.__package_name,
            'dependencies': self.__dependencies,
            'file_size': self.__file_size,
        }

    @property
    def package_name(self) -> str:
        """
        Return the name of the package in which the asset is found.


        :return: The name of the package in which the asset is found.
        """
        return self.__package_name


# class UnrealEngineAsset:
#     def __init__(
#             self,
#             asset_name: str,
#             asset_class_path: str,
#             package_name: str,
#             dependencies: list[str] or None):
#         """
#         Build a new {@link UAsset}.
#
#
#         :param asset_name: The name of the asset without the package.
#
#         :param asset_class_path: The path name of the asset’s class.
#
#         :param package_name: The name of the package in which the asset is
#             found.
#
#         :param dependencies: The list of names of the packages that the asset
#             depends on.
#         """
#         self.__asset_name = asset_name
#         self.__asset_class_path = asset_class_path
#         self.__package_name = package_name
#         self.__dependencies = dependencies
#
#     @property
#     def asset_class_path(self) -> str:
#         """
#         Return the path name of the asset's class.
#
#         Examples:
#
#         ```text
#         /Script/Engine/SkeletalMesh
#         /Script/Engine/Skeleton
#         /Script/Engine/Texture2D
#         ```
#
#         :return: The path name of the asset's class.
#         """
#         return self.__asset_class_path
#
#     @property
#     def asset_name(self) -> str:
#         """
#         Return the name of the asset.
#
#
#         :return: The name of the asset without the package.
#         """
#         return self.__asset_name
#
#     @property
#     def dependencies(self) -> list[str] or None:
#         """
#         Return the list of names of the packages that the asset depends on.
#
#
#         :return: The list of names of the packages that the asset depends on.
#         """
#         return self.__dependencies
#
#     @staticmethod
#     def from_json(payload: any):
#         if isinstance(payload, str):
#             payload = json.loads(payload)
#
#         # @todo: Check data consistency (type, path, list of strings, etc.)
#         return UnrealEngineAsset(
#             payload['asset_name'],
#             payload['asset_class_path'],
#             payload['package_name'],
#             payload['dependencies']
#         )
#
#     def to_json(self) -> any:
#         return {
#             "asset_name": self.__asset_name,
#             "asset_class_path": self.__asset_class_path,
#             "package_name": self.__package_name,
#             "dependencies": self.__dependencies
#         }
#
#     @property
#     def package_name(self) -> str:
#         """
#         Return the name of the package in which the asset is found.
#
#
#         :return: The name of the package in which the asset is found.
#         """
#         return self.__package_name
#
#
# class UAssetFile:
#     FILE_READ_BLOCK_SIZE = 4096
#
#     def __init__(
#             self,
#             asset,
#             file_path_name):
#         if not os.path.exists(file_path_name):
#             error_message = f"The file {file_path_name} of the asset {asset.asset_name} doesn't exist"
#             logging.error(error_message)
#             raise FileNotFoundError(error_message)
#
#         self.__asset = asset
#         self.__file_path_name = file_path_name
#         self.__file_status = None  # This attribute is lazy loaded (cf. property `file_status`)
#         self.__file_checksum = None  # This attribute is lazy loaded (cf. property `file_checksum`)
#
#     def __calculate_file_checksum(self) -> str:
#         sha256_hash = hashlib.sha256()
#
#         with open(self.__file_path_name, 'rb') as fd:
#             # Read and update hash string value in blocks of bytes.
#             for byte_block in iter(lambda: fd.read(self.FILE_READ_BLOCK_SIZE), b''):
#                 sha256_hash.update(byte_block)
#
#         return sha256_hash.hexdigest()
#
#     @property
#     def asset(self) -> UnrealEngineAsset:
#         return self.__asset
#
#     @property
#     def file_checksum(self) -> str:
#         if self.__file_checksum is None:
#             self.__file_checksum = self.__calculate_file_checksum()
#
#         return self.__file_checksum
#
#     @property
#     def file_path_name(self) -> Path:
#         return self.__file_path_name
#
#     @property
#     def file_status(self) -> os.stat_result:
#         """
#         Return the detailed status of the asset file.
#
#
#         :return: The detailed status of the asset file.
#         """
#         if self.__file_status is None:
#             self.__file_status = Path.stat(self.__file_path_name)
#
#         return self.__file_status
#
#     def to_json(self) -> any:
#         payload = self.__asset.to_json()
#         payload['file_size'] = self.file_status.st_size
#         payload['file_checksum'] = self.file_checksum
#         return payload

